from __future__ import annotations

import pprint

from loguru import logger
from pycommence import pycommence_context

from amship.models.cmc_update import make_update_dict
from amship.models.shipment import AmherstShipment, AmherstShipmentRequest
from shipaw.fapi.alerts import Alert, AlertType
from shipaw.fapi.responses import ShipmentResponse


async def try_update_cmc(shipment: AmherstShipment, shipment_response: ShipmentResponse):
    try:
        update_dict = await make_update_dict(shipment, shipment_response)
        logger.info(f'Updating CMC: {update_dict}')
        with pycommence_context(csrname=shipment.row_info.category) as pycmc1:
            pycmc1.update_row(update_dict, row_id=shipment.row_info.id)
        logger.info(
            f'Updated Commence row id {shipment.row_info.id} in {shipment.row_info.category} with:\n{pprint.pformat(update_dict, indent=2)}'
        )

    except ValueError as e:
        msg = f'Error updating Commence: {e}'
        logger.exception(e)
        shipment_response.alerts += Alert(message=msg, type=AlertType.ERROR)


async def cmc_log_callback(request: AmherstShipmentRequest, response: ShipmentResponse):
    request = AmherstShipmentRequest.model_validate(request, from_attributes=True)
    await try_update_cmc(shipment=request.shipment, shipment_response=response)




from functools import cached_property
from typing import TYPE_CHECKING

from pydantic import Field, field_validator

if TYPE_CHECKING:
    from amship.models.amherst_models import AmherstShipableBase
from amship.models.commence_adaptors import AmherstRowInfo
from amship.models.meta import get_table_model
from shipaw.fapi.requests import ShipmentRequest
from shipaw.models.shipment import Shipment


class AmherstShipment(Shipment):
    context: dict = Field(default_factory=dict)

    @field_validator('context', mode='after')
    def val_context(cls, v):
        if 'record' not in v:
            raise ValueError('context must include a "record" key')
        if not hasattr(v['record'], 'row_info') and 'row_info' not in v['record']:
            raise ValueError('context["record"] must include a "row_info" key')
        return v

    @cached_property
    def row_info(self) -> AmherstRowInfo:
        res = self.record_dict.get('row_info')
        return AmherstRowInfo(category=res[0], id=res[1])

    @cached_property
    def record_dict(self):
        return self.context.get('record')

    @cached_property
    def record(self) -> 'AmherstShipableBase':
        model_type = get_table_model(self.row_info.category)
        return model_type.model_validate(self.record_dict)


class AmherstShipmentRequest(ShipmentRequest):
    shipment: AmherstShipment



from __future__ import annotations

from collections.abc import Awaitable, Callable
from enum import StrEnum
from typing import NamedTuple

from fastapi import Query
from pycommence.filters import FilterArray
from pycommence.pycmc_types import Connection, RowFilter

from amship.models.amherst_models import (
    AmherstCustomer,
    AmherstHire,
    AmherstSale,
    AmherstTrial, AmherstShipableBase,
)
from amship.models.commence_adaptors import CategoryName, CustomerAliases, HireAliases, SaleAliases, TrialAliases
from amship.models.filters import (
    CUSOMER_CONNECTION,
    CUSTOMER_ARRAY_LOOSE,
    CUSTOMER_ARRAY_TIGHT,
    HIRE_ARRAY_LOOSE,
    HIRE_ARRAY_TIGHT,
    HIRE_CONNECTION,
    SALE_ARRAY_LOOSE,
    SALE_ARRAY_TIGHT,
    SALE_CONNECTION,
    customer_row_filter_loose,
    hire_row_filter_loose,
    sale_row_filter_loose,
)
from shipaw.fapi.responses import ShipmentResponse
from shipaw.models.shipment import Shipment


class FilterMapPy(NamedTuple):
    loose: RowFilter | None = None
    tight: RowFilter | None = None


class FilterMapCmc(NamedTuple):
    loose: FilterArray | None = None
    tight: FilterArray | None = None


class ConnectionMap(NamedTuple):
    customer: Connection | None = None
    hire: Connection | None = None
    sale: Connection | None = None


class TemplateMap(NamedTuple):
    listing: str
    detail: str


CmcUpdateFunc = Callable[[AmherstShipableBase, Shipment, ShipmentResponse], Awaitable[dict[str, str]]]




class AmherstMap(NamedTuple):
    category: CategoryName
    record_model: type(AmherstShipableBase)
    aliases: type(StrEnum)
    templates: TemplateMap
    # cmc_update_fn: CmcUpdateFuncAgnost | None = make_update_dict_agnost
    connections: ConnectionMap | None = None
    py_filters: FilterMapPy | None = None
    cmc_filters: FilterMapCmc | None = None


class AmherstMaps:
    hire: AmherstMap = AmherstMap(
        category=CategoryName.Hire,
        record_model=AmherstHire,
        aliases=HireAliases,
        templates=TemplateMap(
            listing='order_list.html',
            detail='order_detail.html',
        ),
        connections=ConnectionMap(
            customer=CUSOMER_CONNECTION,
        ),
        py_filters=FilterMapPy(
            loose=hire_row_filter_loose,
            tight=hire_row_filter_loose,
        ),
        cmc_filters=FilterMapCmc(
            loose=HIRE_ARRAY_LOOSE,
            tight=HIRE_ARRAY_TIGHT,
        ),
    )
    sale: AmherstMap = AmherstMap(
        category=CategoryName.Sale,
        record_model=AmherstSale,
        aliases=SaleAliases,
        templates=TemplateMap(
            listing='order_list.html',
            detail='order_detail.html',
        ),
        connections=ConnectionMap(
            customer=CUSOMER_CONNECTION,
        ),
        py_filters=FilterMapPy(
            loose=sale_row_filter_loose,
            tight=sale_row_filter_loose,
        ),
        cmc_filters=FilterMapCmc(
            loose=SALE_ARRAY_LOOSE,
            tight=SALE_ARRAY_TIGHT,
        ),
    )
    customer: AmherstMap = AmherstMap(
        category=CategoryName.Customer,
        record_model=AmherstCustomer,
        aliases=CustomerAliases,
        templates=TemplateMap(
            listing='customer_list.html',
            detail='customer_detail.html',
        ),
        connections=ConnectionMap(
            hire=HIRE_CONNECTION,
            sale=SALE_CONNECTION,
        ),
        py_filters=FilterMapPy(
            loose=customer_row_filter_loose,
            tight=customer_row_filter_loose,
        ),
        cmc_filters=FilterMapCmc(
            loose=CUSTOMER_ARRAY_LOOSE,
            tight=CUSTOMER_ARRAY_TIGHT,
        ),
    )
    trial: AmherstMap = AmherstMap(
        category=CategoryName.Trial,
        record_model=AmherstTrial,
        aliases=TrialAliases,
        templates=TemplateMap(
            listing='customer_list.html',
            detail='customer_detail.html',
        ),
    )
    #
    # repairs: AmherstMap = AmherstMap(
    #     category=CategoryName.Repairs,
    #     record_model=AmherstTrial,
    #     aliases=TrialAliases,
    #     templates=TemplateMap(
    #         listing='customer_list.html',
    #         detail='customer_detail.html',
    #     ),
    #     # cmc_update_fn=make_base_update_dict,
    # )


async def mapper_from_query_csrname(csrname: CategoryName = Query(...)) -> AmherstMap:
    cat = 'trial' if 'trial' in csrname.lower() else csrname.lower()
    return getattr(AmherstMaps, cat)


# async def make_base_update_dict(
#     record: AMHERST_TABLE_MODELS, shipment: SHIPMENT_TYPES, shipment_response: AmherstShipmentResponse
# ) -> dict[str, Any]:
#     """Adds tracking numbers and link."""
#     aliases = await get_alias(record)
#     tracks = await add_tracking_to_list(record, shipment_response)
#     update_package = {aliases.TRACKING_NUMBERS: tracks}
#     return update_package


# async def make_hire_update_dict(
#     record: AmherstHire, shipment: AmherstShipmentOut, shipment_response: AmherstShipmentResponse
# ):
#     aliases = await get_alias(record)
#     update_base = await make_base_update_dict(record, shipment, shipment_response)
#     shipdir = shipment.direction
#     if shipdir in [ShipDirection.INBOUND, ShipDirection.DROPOFF]:
#         extra = {aliases.ARRANGED_IN: 'True', HireAliases.PICKUP_DATE: f'{shipment.shipping_date:%Y-%m-%d}'}
#     elif shipdir == ShipDirection.OUTBOUND:
#         extra = {aliases.ARRANGED_OUT: 'True'}
#     else:
#         raise ValueError(f'Invalid shipment direction: {shipdir}')
#     extra.update(update_base)
#     return extra




from __future__ import annotations

from datetime import date
from enum import StrEnum
from typing import Annotated

import pydantic as _p
from pycommence.pycmc_types import (
    get_cmc_date, RowInfo,
)


class CategoryName(StrEnum):
    Hire = 'Hire'
    Sale = 'Sale'
    Customer = 'Customer'
    Trial = 'Radio Trial'
    # Repairs = 'Repairs'


class ViewCursorName(StrEnum):
    HiresOut = 'Hires Outbound - Paul'
    HiresIn = 'Hires Inbound - Paul'


CursorName = CategoryName | ViewCursorName


class HireStatus(StrEnum):
    BOOKED_IN = 'Booked in'
    PACKED = 'Booked in and packed'
    PARTIALLY_PACKED = 'Partially packed'
    OUT = 'Out'
    RTN_OK = 'Returned all OK'
    RTN_PROBLEMS = 'Returned with problems'
    QUOTE_GIVEN = 'Quote given'
    CANCELLED = 'Cancelled'
    EXTENDED = 'Extended'
    SOLD = 'Sold to customer'


class SaleStatus(StrEnum):
    BOOKED = 'Ordered Ready To Go'
    PACKED = 'Packed'
    SENT = 'Sent'
    WAITING_PAYMENT = 'Waiting For Payment'
    WAITING_OTHER = 'Waiting For Other'
    WAITING_STOCK = 'Waiting For Stock'
    QUOTE = 'Quote Sent'
    LOST_KIT = 'Lost Kit Invoice'
    CANCELLED = 'Cancelled'
    SUPPLIER = 'Sent Direct From Supplier'


AM_DATE = Annotated[
    date | None,
    _p.BeforeValidator(get_cmc_date),
]


class CustomerAliases(StrEnum):
    NAME = 'Name'
    CUSTOMER_NAME = 'Name'

    DELIVERY_CONTACT_NAME = 'Deliv Contact'
    DELIVERY_CONTACT_BUSINESS = 'Deliv Name'
    DELIVERY_CONTACT_PHONE = 'Deliv Telephone'
    DELIVERY_CONTACT_EMAIL = 'Deliv Email'

    DELIVERY_ADDRESS_STR = 'Deliv Address'
    DELIVERY_ADDRESS_PC = 'Deliv Postcode'

    INVOICE_EMAIL = 'Invoice Email'
    ACCOUNTS_EMAIL = 'Accounts Email'

    INVOICE_ADDRESS = 'Invoice Address'
    INVOICE_CONTACT = 'Invoice Contact'
    INVOICE_NAME = 'Invoice Name'
    INVOICE_POSTCODE = 'Invoice Postcode'
    INVOICE_TELEPHONE = 'Invoice Telephone'
    PRIMARY_EMAIL = 'Email'
    DATE_LAST_CONTACTED = 'Date Last Contact'

    HIRES = 'Has Hired Hire'
    SALES = 'Involves Sale'

    TRACK_OUT = 'Track Outbound'
    TRACK_IN = 'Track Inbound'
    TRACKING_NUMBERS = 'Tracking Numbers'


class HireAliases(StrEnum):
    NAME = 'Name'
    CUSTOMER_NAME = 'To Customer'

    DELIVERY_CONTACT_BUSINESS = 'Delivery Name'
    DELIVERY_CONTACT_NAME = 'Delivery Contact'
    DELIVERY_CONTACT_EMAIL = 'Delivery Email'
    DELIVERY_CONTACT_PHONE = 'Delivery Tel'

    DELIVERY_ADDRESS_STR = 'Delivery Address'
    DELIVERY_ADDRESS_PC = 'Delivery Postcode'

    BOXES = 'Boxes'
    SEND_DATE = 'Send Out Date'
    DELIVERY_METHOD = 'Send Method'
    INVOICE = 'Invoice'
    ARRANGED_OUT = 'DB label printed'
    ARRANGED_IN = 'Pickup Arranged'
    TRACK_OUT = 'Track Outbound'
    TRACK_IN = 'Track Inbound'
    TRACKING_NUMBERS = 'Tracking Numbers'
    MISSING_KIT_STR = 'Missing Kit'

    ACTUAL_RETURN_DATE = 'Actual Return Date'
    AERIAL_ADAPT = 'Number Aerial Adapt'
    ALL_ADDRESS = 'All Address'
    BATTERIES = 'Number Batteries'
    BOOKED_DATE = 'Booked Date'
    CASES = 'Number Cases'
    CLIPON_AERIAL = 'Number Clipon Aerial'
    CLOSED = 'Closed'
    DB_LABEL_PRINTED = 'DB label printed'
    DELIVERY_COST = 'Delivery Cost'
    DISCOUNT_DESCRIPTION = 'Discount Description'
    DISCOUNT_PERCENTAGE = 'Discount Percentage'
    DUE_BACK_DATE = 'Due Back Date'
    EM = 'Number EM'
    EMC = 'Number EMC'
    HEADSET = 'Number Headset'
    HEADSET_BIG = 'Number Headset Big'
    ICOM = 'Number Icom'
    ICOM_CAR_LEAD = 'Number ICOM Car Lead'
    ICOM_PSU = 'Number ICOM PSU'
    MAGMOUNT = 'Number Magmount'
    MEGAPHONE = 'Number Megaphone'
    MEGAPHONE_BAT = 'Number Megaphone Bat'
    PACKED_BY = 'Packed By'
    PACKED_DATE = 'Packed Date'
    PACKED_TIME = 'Packed Time'
    PARROT = 'Number Parrot'
    PAYMENT_TERMS = 'Payment Terms'
    PICKUP_ARRANGED = 'Pickup Arranged'
    PICKUP_DATE = 'Pickup Date'
    PURCHASE_ORDER = 'Purchase Order'
    RADIO_TYPE = 'Radio Type'
    RECURRING_HIRE = 'Recurring Hire'
    REFERENCE_NUMBER = 'Reference Number'
    REPEATER = 'Number Repeater'
    REPROGRAMMED = 'Reprogrammed'
    RETURN_NOTES = 'Return Notes'
    SALE_TELEPHONE = 'Delivery Telephone'
    SENDING_STATUS = 'Sending Status'
    SEND_COLLECT = 'Send / Collect'
    SGL_CHARGER = 'Number Sgl Charger'
    SPECIAL_KIT = 'Special Kit'
    STATUS = 'Status'
    UHF = 'Number UHF'
    UHF_6WAY = 'Number UHF 6-way'
    UNPACKED_BY = 'Unpacked by'
    UNPACKED_DATE = 'Unpacked Date'
    UNPACKED_TIME = 'Unpacked Time'
    VHF = 'Number VHF'
    VHF_6WAY = 'Number VHF 6-way'
    WAND = 'Number Wand'
    WAND_BAT = 'Number Wand Battery'
    WAND_CHARGER = 'Number Wand Charger'


class SaleAliases(StrEnum):
    CUSTOMER_NAME = 'To Customer'
    NAME = 'Name'

    DELIVERY_CONTACT_BUSINESS = 'Delivery Name'
    DELIVERY_CONTACT_NAME = 'Delivery Contact'
    DELIVERY_CONTACT_EMAIL = 'Delivery Email'
    DELIVERY_CONTACT_PHONE = 'Delivery Telephone'

    DELIVERY_ADDRESS_STR = 'Delivery Address'
    DELIVERY_ADDRESS_PC = 'Delivery Postcode'

    BOXES = 'Boxes'
    SEND_DATE = 'Date Ordered'
    DELIVERY_METHOD = 'Delivery Method'
    INVOICE = 'Invoice'
    ARRANGED_OUT = 'DB label printed'
    ARRANGED_IN = 'Pickup Arranged'
    TRACK_OUT = 'Track Outbound'
    TRACK_IN = 'Track Inbound'
    TRACKING_NUMBERS = 'Tracking Numbers'

    LOST_EQUIPMENT = 'Lost Equipment'
    STATUS = 'Status'
    DATE_ORDERED = 'Date Ordered'
    DATE_SENT = 'Date Sent'
    INVOICE_TERMS = 'Invoice Terms'
    PURCHASE_ORDER_PRINT = 'Purchase Order Print'
    PURCHASE_ORDER = 'Purchase Order'
    ITEMS_ORDERED = 'Items Ordered'
    SERIAL_NUMBERS = 'Serial Numbers'
    ORDER_PACKED_BY = 'Order Packed By'
    ORDER_TAKEN_BY = 'Order Taken By'
    OUTBOUND_ID = 'Outbound ID'
    NOTES = 'Notes'
    DELIVERY_NOTES = 'Delivery Notes'
    INVOICE_ADDRESS = 'Invoice Address'
    INVOICE_CONTACT = 'Invoice Contact'
    INVOICE_EMAIL = 'Invoice Email'
    INVOICE_NAME = 'Invoice Name'
    INVOICE_POSTCODE = 'Invoice Postcode'
    INVOICE_TELEPHONE = 'Invoice Telephone'


class TrialAliases(StrEnum):
    NAME = 'Name'
    CUSTOMER_NAME = 'Involves Customer'

    DELIVERY_CONTACT_BUSINESS = 'Trial Name'
    DELIVERY_CONTACT_NAME = 'Trial Contact'
    DELIVERY_CONTACT_EMAIL = 'Trial Email'
    DELIVERY_CONTACT_PHONE = 'Trial Telephone'

    DELIVERY_ADDRESS_STR = 'Trial Address'
    DELIVERY_ADDRESS_PC = 'Trial Postcode'

    INVOICE = 'Our Invoice'
    # ARRANGED_OUT = 'DB label printed'
    ARRANGED_IN = 'Pickup Arranged'
    TRACK_OUT = 'Track Outbound'
    TRACK_IN = 'Track Inbound'
    TRACKING_NUMBERS = 'Tracking Numbers'

    STATUS = 'Status'


def get_alias(alias_cls: StrEnum, field_name: str) -> str:
    field_name = field_name.upper()
    if hasattr(alias_cls, field_name):
        return getattr(alias_cls, field_name).value
    return field_name


class RadioType(StrEnum):
    HYTERA = 'Hytera Digital'
    KIRISUN = 'Kirisun UHF'


def replace_curly_apostrophe(value: str) -> str:
    return value.replace('’', "'") if isinstance(value, str) else value


def replace_noncompliant_apostrophes(value: str) -> str:
    if isinstance(value, str):
        for char in NONCOMPLIANT_APOSTROPHES:
            value = value.replace(char, "'")
    return value


def alias_generator_generic(field_name: str, cls) -> str:
    return cls.aliases[field_name.upper()].value


def customer_alias_generator(field_name: str) -> str:
    try:
        return CustomerAliases[field_name.upper()].value
    except KeyError:
        # logger.warning(f'No Customer alias found for field name: {field_name}')
        return field_name


def sale_alias_generator(field_name: str) -> str:
    try:
        return SaleAliases[field_name.upper()].value
    except KeyError:
        # logger.warning(f'No Sale alias found for field name: {field_name}')
        return field_name


def hire_alias_generator(field_name: str) -> str:
    try:
        return HireAliases[field_name.upper()].value
    except KeyError:
        # logger.warning(f'No Hire alias found for field name: {field_name}')
        return field_name


def trial_alias_generator(field_name: str) -> str:
    try:
        return TrialAliases[field_name.upper()].value
    except KeyError:
        # logger.warning(f'No Trial alias found for field name: {field_name}')
        return field_name


NONCOMPLIANT_APOSTROPHES = ['’', '‘', '′', 'ʼ', '´']


def split_shipment_refs_dict_from_str(ref_str: str, max_len: int = 24) -> dict[str, str]:
    reference_numbers = {}

    for i in range(1, 6):
        start_index = (i - 1) * max_len
        end_index = i * max_len
        if start_index < len(ref_str):
            reference_numbers[f'reference_number{i}'] = ref_str[start_index:end_index]
        else:
            break
    return reference_numbers


def split_addr_str2(address: str) -> tuple[list[str], str]:
    addr_lines = address.splitlines()
    town = addr_lines.pop() if len(addr_lines) > 1 else ''

    if len(addr_lines) < 3:
        addr_lines.extend([''] * (3 - len(addr_lines)))
    elif len(addr_lines) > 3:
        addr_lines[2] = ','.join(addr_lines[2:])
        addr_lines = addr_lines[:3]

    used_lines = [_ for _ in addr_lines if _]
    return used_lines, town


class AmherstRowInfo(RowInfo):
    category: CategoryName



"""
Wrap FastAPI app in FlaskWebGUI for desktop application.
Use `Paw Request fork <https://github.com/pawrequest/flaskwebgui>`_ for URL_SUFFIX to dynamically set loading page to the retrieved record
"""

import argparse
import asyncio
import os
from pathlib import Path

import dotenv
from loguru import logger

from amship.models.commence_adaptors import CategoryName


def shipper_cli():
    args = parse_ship_args()
    logger.info(f'starting shipper for {args.category} {args.record_name} with env {args.shipaw_env}')
    os.environ['SHIPAW_ENV'] = str(args.shipaw_env)
    dotenv.load_dotenv(args.shipaw_env)
    from amship.ui_runner import pycommence_shipper  # AFTER SETTING ENVIRONMENT

    asyncio.run(pycommence_shipper(args.category, args.record_name))


def parse_ship_args():
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument('category', type=CategoryName, choices=list(CategoryName))
    arg_parser.add_argument('record_name', type=str)
    arg_parser.add_argument('shipaw_env', type=Path)
    args = arg_parser.parse_args()
    return args




"""Wrap FastAPI app in FlaskWebGUI for desktop application."""

from __future__ import annotations
from flaskwebgui import FlaskUI, close_application
from jinja2.utils import url_quote
from loguru import logger

from amship import app
from amship.models.commence_adaptors import CategoryName


async def run_desktop_ui(url_suffix='', port=8000):
    app.app.starting_url = url_suffix
    try:
        logger.info(f'Running WebFlaskUI @{url_suffix}')
        FlaskUI(
            fullscreen=True,
            app=app.app,
            server='fastapi',
            port=port,
            app_mode=False,
        ).run()
    finally:
        close_application()


async def pycommence_shipper(category: CategoryName, record_name: str):
    url_suffix = await get_pycommence_shipper_url(category, record_name)
    await run_desktop_ui(url_suffix)


async def get_pycommence_shipper_url(category: CategoryName, record_name: str) -> str:
    return (
        f'shipaw/ship_form_am?csrname={url_quote(category)}&pk_value={url_quote(record_name)}&condition=equal&max_rtn=1'
    )


REVIEW_URL = r'/shipaw/order_review_am'
CONFIRM_URL = r'/shipaw/post_confirm_am'



from __future__ import annotations

import functools
from datetime import datetime, timedelta
from enum import StrEnum
from typing import Literal
from collections.abc import Generator

from loguru import logger
from pycommence.filters import ConditionType, ConnectedFieldFilter, FieldFilter, FilterArray, Sort, SortOrder
from pycommence.pycmc_types import Connection, get_cmc_date

from amship.models.commence_adaptors import CustomerAliases, HireAliases, SaleAliases

CUTOFF_DATE = (datetime.now() - timedelta(days=300)).date()
FilterVariant = Literal['loose', 'tight']


def customer_row_filter_loose(rowgen: Generator[dict[str, str], None, None]) -> Generator[dict[str, str], None, None]:
    for row in rowgen:
        if contacted := row.get(CustomerAliases.DATE_LAST_CONTACTED):
            datey = get_cmc_date(contacted)
            if not datey:
                logger.warning(f'Invalid date for {CustomerAliases.DATE_LAST_CONTACTED}: {contacted}')
            if datey < CUTOFF_DATE:
                continue
            yield row


def order_row_filter_loose(
    aliases: type[StrEnum], rowgen: Generator[dict[str, str], None, None]
) -> Generator[dict[str, str], None, None]:
    for row in rowgen:
        # if hasattr(aliases, 'SEND_DATE'):
        #     if datey := row.get(aliases.SEND_DATE):
        #         send_date = get_cmc_date(datey)
        #         if send_date < CUTOFF_DATE:
        #             continue
        yield row


def hire_row_filter_loose(rowgen: Generator[dict[str, str], None, None]) -> Generator[dict[str, str], None, None]:
    yield from order_row_filter_loose(HireAliases, rowgen)


def sale_row_filter_loose(rowgen: Generator[dict[str, str], None, None]) -> Generator[dict[str, str], None, None]:
    yield from order_row_filter_loose(SaleAliases, rowgen)


HIRE_ARRAY_TIGHT = FilterArray(
    filters={
        1: FieldFilter(column=HireAliases.STATUS, condition=ConditionType.CONTAIN, value='Booked'),
        2: FieldFilter(column=HireAliases.SEND_DATE, condition=ConditionType.AFTER, value='one week ago'),
        3: FieldFilter(column=HireAliases.SEND_DATE, condition=ConditionType.BEFORE, value='one week from today'),
        4: FieldFilter(column=HireAliases.ARRANGED_OUT, condition=ConditionType.NOT),
    },
    sorts=[Sort(column=HireAliases.SEND_DATE, order=SortOrder.ASC)],
    # sorts=[
    #     (HireAliases.SEND_DATE, SortOrder.ASC),
    # ],
    logics=['And', 'And', 'And'],
)

HIRE_ARRAY_LOOSE = FilterArray(
    filters={
        1: FieldFilter(column=HireAliases.SEND_DATE, condition=ConditionType.AFTER, value='six month ago'),
        2: FieldFilter(column=HireAliases.SEND_DATE, condition=ConditionType.BEFORE, value='three month from today'),
        # 2: FieldFilter(column=HireAliases.SEND_DATE, condition=ConditionType.BETWEEN, value='six months ago, three months from today'),
    },
    sorts=[Sort(column=HireAliases.SEND_DATE, order=SortOrder.ASC)],
    logics=['And'],
)


SALE_ARRAY_TIGHT = FilterArray(
    filters={
        1: FieldFilter(column=SaleAliases.DATE_ORDERED, condition=ConditionType.AFTER, value='one month ago'),
    },
    sorts=[Sort(column=SaleAliases.DATE_ORDERED, order=SortOrder.DESC)],
    # sorts=[(SaleAliases.DATE_ORDERED, SortOrder.DESC)],
)

SALE_ARRAY_LOOSE = SALE_ARRAY_TIGHT

HIRE_CONNECTION = Connection(name='Has Hired', category='Hire', column='Name')
SALE_CONNECTION = Connection(name='Involves', category='Sale', column='Name')
CUSOMER_CONNECTION = Connection(name='To', category='Customer', column='Name')


@functools.lru_cache
def customer_array_tight():
    hire_fils = HIRE_ARRAY_TIGHT.filters.values()
    customer_hire_filters = [ConnectedFieldFilter.from_fil(f, HIRE_CONNECTION) for f in hire_fils]
    sale_fils = SALE_ARRAY_TIGHT.filters.values()
    customer_sale_filters = [ConnectedFieldFilter.from_fil(f, SALE_CONNECTION) for f in sale_fils]

    return FilterArray(
        filters={
            1: customer_hire_filters[0],
            2: customer_hire_filters[1],
            3: customer_hire_filters[2],
            4: customer_hire_filters[3],
            5: customer_sale_filters[0],
        },
        logics=['And', 'And', 'And', 'Or'],
        # logics=hire_logics + ['Or'] + sale_logics,
    )


CUSTOMER_ARRAY_TIGHT = customer_array_tight()

CUSTOMER_ARRAY_LOOSE = FilterArray(
    filters={
        1: FieldFilter(column=CustomerAliases.DATE_LAST_CONTACTED, condition=ConditionType.AFTER, value='one year ago'),
    },
    logics=['And'],
)
#
# @functools.lru_cache
# def get_customer_filter2(pk_value: str | None = None):
#     customer_hire_connection = Connection2(name='Has Hired', category='Hire', column='Name')
#     customer_sale_connection = Connection2(name='Involves', category='Sale', column='Name')
#     customer_hire_filters = [
#         ConnectedFieldFilter.from_fil(f, customer_hire_connection) for f in get_hire_filter().filters.values()
#     ]
#     customer_sale_filters = [
#         ConnectedFieldFilter.from_fil(f, customer_sale_connection) for f in get_sale_filter().filters.values()
#     ]
#     assert len(customer_hire_filters) == 3
#     assert len(customer_sale_filters) == 1
#
#     if pk_value:
#         res = FilterArray(
#             filters={
#                 1: customer_hire_filters[0],
#                 2: customer_hire_filters[1],
#                 3: customer_hire_filters[2],
#                 4: FieldFilter(column=HireAliases.NAME, condition=ConditionType.CONTAIN, value=pk_value),
#                 5: FieldFilter(column=SaleAliases.NAME, condition=ConditionType.CONTAIN, value=pk_value),
#                 6: customer_sale_filters[0],
#             },
#             logics=['And', 'And', 'And', 'Or', 'And'],
#         )
#     else:
#         res = FilterArray(
#             filters={
#                 1: customer_hire_filters[0],
#                 2: customer_hire_filters[1],
#                 3: customer_sale_filters[0],
#             },
#             logics=['And', 'Or'],
#         )
#     return res


# @functools.lru_cache
# def get_filter_array(
#     filtername: FilterName,
#     pk_value: str | None = None,
# ) -> FilterArray:
#     match filtername.lower():
#         case 'hire':
#             return get_hire_filter(pk_value)
#         case 'sale':
#             return get_sale_filter(pk_value)
#         case 'customer':
#             return get_customer_filter(pk_value)
#         case _:
#             raise ValueError(f'Unknown cursor name {filtername}')



from fastapi import APIRouter, Depends
from loguru import logger
from starlette.requests import Request
from starlette.responses import HTMLResponse

from amship.back.backend_pycommence import pycommence_get_one
from amship.back.callbacks import cmc_log_callback
from amship.models.amherst_models import AmherstHire, AmherstShipableBase
from shipaw.fapi.routes_html import shipping_form
from shipaw.fapi.alerts import Alert, AlertType, Alerts

router = APIRouter()


@router.get('/ship_form_am', response_class=HTMLResponse)
async def get_shipping_form(
    request: Request,
    record: AmherstShipableBase = Depends(pycommence_get_one),
) -> HTMLResponse:
    alerts: Alerts = request.app.alerts
    if isinstance(record, AmherstHire) and 'parcelforce' not in record.delivery_method.lower():
        msg = f'"Parcelforce" not in delivery_method: {record.delivery_method}'
        logger.warning(msg)
        alerts += Alert(message=msg, type=AlertType.WARNING)

    shipment = record.shipment()
    request.app.callback = cmc_log_callback

    return await shipping_form(request=request, shipment=shipment)




"""
Module for backend integration with PyCommence data sources.

Provides async functions for querying, fetching, and searching records
using PyCommence, with support for pagination and filtering.
"""

from __future__ import annotations

from collections.abc import AsyncGenerator

from fastapi import Depends, Query
from loguru import logger
from pycommence.cursor import RESULTS_GENERATOR
from pycommence.exceptions import PyCommenceNotFoundError
from pycommence.pycmc_types import RowData
from starlette.exceptions import HTTPException
from pycommence import MoreAvailable, PyCommence, pycommence_context, pycommences_context

from amship.back.backend_search_paginate import SearchRequest, SearchResponse, MoreAvailableFront
from amship.models.commence_adaptors import CursorName, AmherstRowInfo
from amship.models.maps import CategoryName, mapper_from_query_csrname
from amship.models.meta import TABLE_REGISTER, get_table_model
from amship.models.amherst_models import AmherstShipableBase


async def pycmc_f_query(
    csrname: CursorName = Query(...),
) -> AsyncGenerator[PyCommence, None]:
    with pycommence_context(csrname=csrname) as pycmc:
        yield pycmc


async def pycmcs_f_query(
    csrnames: list[CategoryName] = Query(...),
) -> AsyncGenerator[PyCommence, None]:
    with pycommences_context(csrnames=csrnames) as pycmc:
        yield pycmc


async def pycommence_gather(
    pycmc: PyCommence,
    q: SearchRequest,
) -> tuple[list[AmherstShipableBase], MoreAvailable | None]:
    """
    Gather records from PyCommence based on the provided search request.
    Add MoreAvailable if q has pagination and there are more records to fetch.
    """

    model_type = get_table_model(q.csrname)
    logger.warning('GATHERING')
    more = None
    records = []
    for row in await generate_pycommence_rowdata(pycmc, q):
        if isinstance(row, MoreAvailable):
            more = MoreAvailableFront(n_more=row.n_more, json_link=q.next_q_str_json, html_link=q.next_q_str)
            break
        records.append(convert_pycommence_rowdata(model_type, row))
    return records, more


def convert_pycommence_rowdata[T: type[AmherstShipableBase]](input_type: T, result: RowData) -> T:
    mod = input_type(row_info=result.row_info, **result.data)
    return mod.model_validate(mod)


async def generate_pycommence_rowdata(
    pycmc: PyCommence,
    q: SearchRequest,
) -> RESULTS_GENERATOR:
    logger.debug(f'Generating rows for {q.csrname}')
    mapper = await mapper_from_query_csrname(csrname=q.csrname)
    fil_array = await q.filter_array()
    row_filter_fn = mapper.py_filters[q.py_filter_i] if q.py_filter_i else None
    return pycmc.read_rows(
        csrname=q.csrname,
        pagination=q.pagination,
        row_filter=row_filter_fn,
        filter_array=fil_array,
    )


async def pycommence_fetch(
    q: SearchRequest,
    pycmc: PyCommence,
) -> AmherstShipableBase | None:
    row_id = None
    if q.row_id:
        row_id = q.row_id
    elif q.pk_value:
        try:
            row_id = pycmc.csr(q.csrname).pk_to_id(q.pk_value)
        except PyCommenceNotFoundError:
            ...
    if row_id:
        row = pycmc.read_row(csrname=q.csrname, row_id=row_id)
        # mapper = await mapper_from_query_csrname(csrname=q.csrname)
        # res = mapper.record_model(row_info=row.row_info, **row.data)
        model_type = TABLE_REGISTER.get(q.csrname)
        res = model_type(row_info=row.row_info, **row.data)
        return res
    return None


async def pycommence_fetch_f_info(
    row_info: AmherstRowInfo,
) -> AmherstShipableBase | None:
    with pycommence_context(csrname=row_info.category) as pycmc:
        row = pycmc.read_row(csrname=row_info.category, row_id=row_info.id).data
    model_type = get_table_model(row_info.category)
    return model_type.model_validate(row)


async def pycommence_search(
    q: SearchRequest = Depends(SearchRequest.from_query),
    pycmc: PyCommence = Depends(pycmc_f_query),
) -> SearchResponse:
    if record := await pycommence_fetch(q, pycmc):
        records, more = [record], None
    else:
        records, more = await pycommence_gather(pycmc=pycmc, q=q)

    resp = SearchResponse(records=records, more=more, search_request=q)

    if q.max_rtn and resp.length > q.max_rtn:
        raise HTTPException(
            status_code=404,
            detail=f'Too many items found: Specified {q.max_rtn} rows and returned {resp.length}',
        )
    return resp


async def pycommence_get_one(
    search_request: SearchRequest = Depends(SearchRequest.from_query),
    pycmc: PyCommence = Depends(pycmc_f_query),
) -> AmherstShipableBase:
    search_request.max_rtn = 1
    res = await pycommence_fetch(search_request, pycmc)
    if not res:
        logger.warning('No Results')
    return res




from typing import TYPE_CHECKING

from loguru import logger

from amship.models.commence_adaptors import CategoryName

if TYPE_CHECKING:
    from amship.models.amherst_models import AmherstShipableBase

TABLE_REGISTER: dict[str, type['AmherstShipableBase']] = {}


def register_table(cls: type['AmherstShipableBase']) -> type['AmherstShipableBase']:
    TABLE_REGISTER[str(cls.category)] = cls
    logger.debug(f'Registered table model: {cls.category}')
    return cls


def get_table_model(csrname: CategoryName) -> type['AmherstShipableBase'] | None:
    res = TABLE_REGISTER.get(csrname)
    if not res:
        logger.warning(f'No table model found for csrname: {csrname}')
    return res





import contextlib

from fastapi import FastAPI, responses
from fastapi.exceptions import RequestValidationError
from shipaw.config import ShipawSettings
from shipaw.fapi.alerts import Alerts
from shipaw.fapi.app import request_validation_exception_handler
from shipaw.fapi.routes_api import router as shipaw_json_router
from shipaw.fapi.routes_html import router as shipaw_html_router
from starlette.requests import Request
from starlette.responses import HTMLResponse, RedirectResponse
from starlette.staticfiles import StaticFiles

from amship.back.ship_routes import router as ship_router


@contextlib.asynccontextmanager
async def lifespan(app_: FastAPI):
    try:
        # set_pf_env()
        # pythoncom.CoInitialize()
        # with sqm.Session(am_db.ENGINE) as session:
        #     pf_shipper = ELClient()
        #     populate_db_from_cmc(session, pf_shipper)
        yield

    finally:
        # pythoncom.CoUninitialize()

        ...


app = FastAPI(lifespan=lifespan)
app.mount('/static', StaticFiles(directory=str(ShipawSettings.from_env().static_dir)), name='static')
# app.mount('/static', StaticFiles(directory=str(amherst_settings().static_dir)), name='static')
app.include_router(ship_router, prefix='/shipaw')
app.include_router(shipaw_json_router, prefix='/shipaw/api')
app.include_router(shipaw_html_router, prefix='/shipaw')
# app.ship_live = pf_config.pf_sett().ship_live
app.alerts = Alerts.empty()


@app.exception_handler(RequestValidationError)
async def request_exception_handler(request: Request, exc: RequestValidationError):
    return await request_validation_exception_handler(request, exc)


@app.get('/robots.txt', response_class=responses.PlainTextResponse)
async def robots_txt() -> str:
    return 'User-agent: *\nAllow: /'


@app.get('/favicon.ico', include_in_schema=False)
async def favicon_ico():
    return responses.RedirectResponse(url='/static/favicon.svg')


@app.get('/', response_class=RedirectResponse)
async def startup(
    request: Request,
):
    if not hasattr(request.app, 'starting_url'):
        raise ValueError('App has no attribute starting_url')
    return RedirectResponse(request.app.starting_url)




# from __future__ import annotations
#
# import os
# from functools import lru_cache
# from pathlib import Path
#
# import pydantic as _p
# from dotenv import dotenv_values, load_dotenv
# from pawlogger import get_loguru
# from pydantic import computed_field
# from pydantic_settings import BaseSettings, SettingsConfigDict
#
#
# # def load_env_index(envs_index: Path) -> None:
# #     load_dotenv(envs_index)
# #     for env in ('SHIPAW_ENV', 'AMHERST_ENV'):
# #         if not os.getenv(env):
# #             raise ValueError(f'Environment variable {env} not set in {envs_index}')
# #         if not Path(os.getenv(env)).exists():
# #             raise ValueError(f'Environment variable {env} points to non-existent file {os.getenv(env)}')
# #         getlog()
#
#
# # def getlog():
# #     env_values = dotenv_values(Path(os.getenv('AMHERST_ENV')))
# #     log_dir = Path(env_values.get('LOG_DIR'))
# #     log_dir.mkdir(parents=True, exist_ok=True)
# #     log_file = log_dir / 'amherst.log'
# #     log_file.touch(exist_ok=True)
# #     log_level = env_values.get('LOG_LEVEL') or 'DEBUG'
# #     get_loguru(log_file=log_file, profile='local', level=log_level)
#
#
# # def load_env() -> Path:
# #     ei = Path(os.environ.get('ENV_INDEX'))
# #     if not ei or not ei.exists():
# #         raise ValueError('ENV_INDEX not set or does not exist')
# #     load_env_index(ei)
# #     print(f'Loading env index from {ei}')
# #     amherst_env = Path(os.getenv('AMHERST_ENV'))
# #     print(f'Loading Amherst Settings from {amherst_env}')
# #     return amherst_env
#
#
# # class Settings(BaseSettings):
# #     log_dir: Path
# #     log_level: str = 'DEBUG'
# #
# #     @computed_field
# #     @property
# #     def log_file(self) -> Path:
# #         return self.log_dir / 'amherst.log'
# #
# #     @computed_field
# #     @property
# #     def ndjson_file(self) -> Path:
# #         return self.log_dir / 'amherst.ndjson'
# #
# #     @_p.model_validator(mode='after')
# #     def create_log_files(self):
# #         self.log_dir.mkdir(parents=True, exist_ok=True)
# #         for v in (self.log_file, self.ndjson_file):
# #             if not v.exists():
# #                 v.touch()
# #         return self
# #
# #     model_config = SettingsConfigDict(env_file=load_env())
#
#
# # @lru_cache
# # def amherst_settings() -> Settings:
# #     res = Settings()
# #     return res
#



from abc import ABC
from datetime import date
from os import PathLike
from typing import ClassVar

from pycommence.pycmc_types import RowInfo
from pydantic import BaseModel, ConfigDict, Field, field_validator

from amship.models.commence_adaptors import (
    AM_DATE,
    CategoryName,
    HireStatus,
    SaleStatus,
    customer_alias_generator,
    hire_alias_generator,
    replace_noncompliant_apostrophes,
    sale_alias_generator,
    split_addr_str2,
    trial_alias_generator,
)
from amship.models.meta import register_table
from amship.models.shipment import AmherstShipment
from shipaw.models.address import Address as AddressAgnost, Contact, FullContact
from shipaw.models.ship_types import ShipDirection


class AmherstShipableBase(BaseModel, ABC):
    category: ClassVar[CategoryName]
    model_config = ConfigDict(
        populate_by_name=True,
        use_enum_values=True,
    )

    def alias_lookup(self, field_name: str) -> str:
        try:
            return self.model_fields[field_name].alias
        except KeyError:
            return field_name

    @field_validator('*', mode='before')
    def preprocess_strings(cls, value):
        return replace_noncompliant_apostrophes(value)

    row_info: RowInfo
    # amherst common fieldnames fields
    name: str = Field(..., alias='Name')
    tracking_numbers: str = Field('', alias='Tracking Numbers')
    track_out: str | None = None
    track_in: str | None = None

    # mandatory fields
    customer_name: str
    delivery_contact_name: str
    delivery_contact_business: str
    delivery_contact_phone: str
    delivery_contact_email: str
    delivery_address_str: str
    delivery_address_pc: str

    # optional fields with default
    send_date: AM_DATE = date.today()
    boxes: int = 1


    @field_validator('send_date', mode='after')
    def validate_send_date(cls, v: AM_DATE) -> date:
        if v is None or v < date.today():
            return date.today()
        return v

    @property
    def full_contact(self) -> FullContact:
        addrlines, town = split_addr_str2(self.delivery_address_str)
        return FullContact(
            contact=Contact(
                contact_name=self.delivery_contact_name,
                mobile_phone=self.delivery_contact_phone,
                email_address=self.delivery_contact_email,
            ),
            address=AddressAgnost(
                address_lines=addrlines,
                town=town,
                postcode=self.delivery_address_pc,
                business_name=self.delivery_contact_business,
            ),
        )

    def shipment(self, direction: ShipDirection = ShipDirection.OUTBOUND) -> 'AmherstShipment':
        return AmherstShipment(
            recipient=self.full_contact,
            boxes=self.boxes,
            shipping_date=self.send_date,
            direction=direction,
            reference=self.customer_name,
            context={'record': self},
        )

    # def shipment1(self, direction: ShipDirection = ShipDirection.OUTBOUND) -> Shipment:
    #     return Shipment(
    #         recipient=self.full_contact,
    #         boxes=self.boxes,
    #         shipping_date=self.send_date,
    #         direction=direction,
    #         reference=self.customer_name,
    #     )


class AmherstOrderBase(AmherstShipableBase, ABC):
    # order fields common
    status: str
    arranged_in: str | None = None
    arranged_out: str | None = None
    invoice: PathLike | None = None

    # order fields optional
    date_sent: AM_DATE | None = None
    booking_date: AM_DATE | None = None


@register_table
class AmherstCustomer(AmherstShipableBase):
    model_config = ConfigDict(
        alias_generator=customer_alias_generator,
        # alias_generator=lambda field_name: CustomerAliases[field_name.upper()].value,
    )
    category: ClassVar[CategoryName] = 'Customer'

    # customer fields
    invoice_email: str
    accounts_email: str
    hires: str = ''
    sales: str = ''


@register_table
class AmherstHire(AmherstOrderBase):
    # optional overrides master
    # aliases: ClassVar[StrEnum] = HireAliases
    model_config = ConfigDict(
        alias_generator=hire_alias_generator,
    )
    category: ClassVar[CategoryName] = 'Hire'

    boxes: int = 1
    delivery_contact_phone: str
    send_date: AM_DATE = date.today()

    delivery_method: str | None = None

    # optional overrides order
    status: HireStatus
    date_sent: AM_DATE | None = None
    booking_date: AM_DATE | None = None

    # hire fields
    missing_kit_str: str | None = None
    due_back_date: AM_DATE
    return_notes: str | None = None


@register_table
class AmherstSale(AmherstOrderBase):
    model_config = ConfigDict(
        alias_generator=sale_alias_generator,
    )
    category: ClassVar[CategoryName] = 'Sale'

    delivery_method: str | None = None

    # optional overrides order
    status: SaleStatus
    booking_date: AM_DATE = date.today()
    date_sent: AM_DATE | None = None

    # sale fields
    lost_equipment: str | None = None
    invoice_terms: str | None = None
    purchase_order: str | None = None
    items_ordered: str | None = None
    serial_numbers: str | None = None
    delivery_notes: str | None = None
    notes: str | None = None


@register_table
class AmherstTrial(AmherstOrderBase):
    # aliases: ClassVar[StrEnum] = TrialAliases
    model_config = ConfigDict(
        alias_generator=trial_alias_generator,
    )
    category: ClassVar[CategoryName] = 'Trial'


#
# class AmherstRepairs(AmherstOrderBase):
#     # category:CategoryName = CategoryName.Trial
#     customer_name: str = Field(..., alias='For Customer')
#     delivery_contact_name: str = Field(..., alias='Trial Contact')
#     delivery_contact_business: str = Field(..., alias='Trial Name')
#     delivery_contact_phone: str = Field(..., alias='Trial Telephone')
#     delivery_contact_email: str = Field(..., alias='Trial Email')
#     delivery_address_str: str = Field(..., alias='Trial Address')
#     delivery_address_pc: str = Field(..., alias='Trial Postcode')
#     tracking_numbers: str = Field('', alias='Tracking Numbers')
#
#     invoice: str = Field('', alias='Our Invoice')


AMHERST_ORDER_MODELS = AmherstHire | AmherstSale | AmherstTrial





from __future__ import annotations

from collections.abc import Awaitable, Callable
from datetime import date
from typing import Any

from amship.models.amherst_models import (
    AmherstHire,
    AmherstShipableBase,
)
from amship.models.shipment import AmherstShipment
from shipaw.fapi.responses import ShipmentResponse
from shipaw.models.ship_types import ShipDirection

CmcUpdateFuncAgnost = Callable[
    [AmherstShipableBase, AmherstShipment, ShipmentResponse], Awaitable[dict[str, str]]
]


def split_com_sep_str_field(record, fieldname: str) -> list[str]:
    data_s: str = getattr(record, fieldname, None)
    data_l = data_s.split(',') if data_s else []
    return data_l


def add_to_com_sep_str_field(data: list, value) -> str:
    data.append(value)
    return ','.join(data)


async def add_tracking_to_list(record: AmherstShipableBase, resp: ShipmentResponse) -> str:
    tracks = split_com_sep_str_field(record, 'tracking_numbers')
    return add_to_com_sep_str_field(tracks, resp.shipment_num)


async def make_update_dict(shipment: AmherstShipment, shipment_response: ShipmentResponse) -> dict[str, Any]:
    """Adds tracking numbers and link."""
    record = shipment.record
    update_package = await cmc_update_dict(shipment, shipment_response)
    if isinstance(record, AmherstHire):
        extra = await cmc_update_dict_hire(shipment)
        update_package.update(extra)
    return update_package


async def cmc_update_dict(shipment: AmherstShipment, shipment_response: ShipmentResponse):
    record = shipment.record
    shipdir = shipment.direction
    tracks = await add_tracking_to_list(record, shipment_response)
    update_package = {record.alias_lookup('tracking_numbers'): tracks}
    if shipdir in [ShipDirection.INBOUND, ShipDirection.DROPOFF]:
        update_package.update({record.alias_lookup('track_in'): shipment_response.tracking_link})
    elif shipdir == ShipDirection.OUTBOUND:
        update_package.update({record.alias_lookup('track_out'): shipment_response.tracking_link})
    else:
        raise ValueError(f'Invalid shipment direction: {shipdir}')
    return update_package


async def cmc_update_dict_hire(shipment: AmherstShipment):
    record = shipment.record
    if not isinstance(record, AmherstHire):
        raise ValueError('Record is not an AmherstHire')
    shipdir = shipment.direction
    if shipdir in [ShipDirection.INBOUND, ShipDirection.DROPOFF]:
        return await cmc_update_dict_hire_in(record, shipment)
    elif shipdir == ShipDirection.OUTBOUND:
        return await cmc_update_dict_hire_out(record)
    else:
        raise ValueError(f'Invalid shipment direction: {shipdir}')


async def cmc_update_dict_hire_in(record: AmherstHire, shipment: AmherstShipment):
    ret_notes = f'{date.today().strftime('%d/%m')}: pickup arranged for {shipment.shipping_date.strftime('%d/%m')}\r\n{record.return_notes}'
    return {
        record.alias_lookup('arranged_in'): 'True',
        record.alias_lookup('pickup_date'): f'{shipment.shipping_date:%Y-%m-%d}',
        record.alias_lookup('return_notes'): ret_notes,
    }


async def cmc_update_dict_hire_out(record: AmherstHire):
    return {
        record.alias_lookup('arranged_out'): 'True',
    }




""" Backend search request and response models with pagination support. """
from __future__ import annotations

import dataclasses
from collections.abc import Sequence
from typing import Self

from fastapi import Depends, Query
from pycommence.filters import ConditionType, ConnectedFieldFilter, FieldFilter, FilterArray
from pycommence.pycmc_types import MoreAvailable, Pagination as _Pagination
from pydantic import BaseModel, Field, model_validator

from amship.models.amherst_models import AmherstShipableBase
from amship.models.commence_adaptors import CursorName, CustomerAliases
from amship.models.maps import AmherstMap, CategoryName, mapper_from_query_csrname

PAGE_SIZE = 50


async def get_condition(condition: str = Query('')) -> ConditionType:
    return getattr(ConditionType, condition.upper(), ConditionType.CONTAIN)


class Pagination(_Pagination):
    @classmethod
    def from_query(cls, limit: int = Query(PAGE_SIZE), offset: int = Query(0)) -> Self:
        return cls(limit=limit, offset=offset)


class SearchRequest(BaseModel):
    csrname: CursorName | None = None
    csrnames: list[CursorName] | None = None
    row_id: str | None = None
    pk_value: str | None = None
    customer_name: str | None = None
    customer_names: list[str] | None = Field(default_factory=list)
    condition: ConditionType = ConditionType.CONTAIN
    max_rtn: int | None = None
    pagination: Pagination | None = Pagination()
    cmc_filter_i: int = 0
    py_filter_i: int = 0

    @model_validator(mode='after')
    def cursornames(self):
        if not self.csrname and not self.csrnames:
            raise ValueError('No csrname or csrnames provided')
        if self.csrname and not self.csrnames:
            self.csrnames = [self.csrname]
        return self

    def __str__(self):
        return (
            f'Csr: {self.csrname if self.csrname else ', '.join(self.csrnames)}'
            f'{' | pk=:' + self.pk_value if self.pk_value else ''}'
            f'{' | row_id=:' + self.row_id if self.row_id else ''}'
            f'{' | customer_name="' + self.customer_name + '"' if self.customer_name else ''}'
            f'{' | cmc_filter_i=' + str(self.cmc_filter_i) if self.cmc_filter_i else ''}'
            f'{' | py_filter_i=' + str(self.py_filter_i) if self.py_filter_i else ''}'
            f'{' | ' + str(self.pagination) if self.pagination else ''}'
        )

    @property
    def q_str(self):
        return self.q_str_paginate()

    @property
    def query_str_json(self):
        return self.q_str_paginate(api=True)

    @property
    def next_q_str(self):
        return self.q_str_paginate(self.pagination.next_page()) if self.pagination else None

    @property
    def next_q_str_json(self):
        return self.q_str_paginate(self.pagination.next_page(), api=True) if self.pagination else None

    def q_str_paginate(self, pagination: Pagination = None, api: bool = False):
        pagination = pagination or self.pagination
        qstr = '/api' if api else ''
        qstr += f'/search?csrname={self.csrname}'
        for attr in [
            'condition',
            'max_rtn',
            'cmc_filter_i',
            'py_filter_i',
            'pk_value',
            'row_id',
            'customer_id',
            'customer_name',
        ]:
            if val := getattr(self, attr):
                qstr += f'&{attr}={val}'
        if pagination:
            if pagination.limit:
                qstr += f'&limit={pagination.limit}'
            if pagination.offset:
                qstr += f'&offset={pagination.offset}'
        return qstr

    def next_request(self):
        return self.model_copy(update={'pagination': self.pagination.next_page()})

    def prev_request(self):
        return self.model_copy(update={'pagination': self.pagination.prev_page()})

    # @classmethod
    # @resolve_row_id
    # def from_id_or_pk(
    #     cls,
    #     csrname: CursorName = Query(...),
    #     pk: str = Query(''),
    #     row_id: str = Query(None),
    # ):
    #     return cls(
    #         csrname=csrname,
    #         pk_value=pk,
    #         row_id=row_id,
    #     )

    @classmethod
    def from_query(
        cls,
        csrname: CategoryName = Query(None),
        csrnames: list[CategoryName] = Query(None),
        pk_value: str = Query(''),
        pagination: Pagination = Depends(Pagination.from_query),
        condition: ConditionType = Depends(get_condition),
        max_rtn: int = Query(None),
        row_id: str = Query(None),
        customer_name: str = Query(None),
        py_filter_i: int = Query(0),
        cmc_filter_i: int = Query(0),
    ):
        return cls(
            csrname=csrname,
            csrnames=csrnames,
            pagination=pagination,
            pk_value=pk_value,
            condition=condition,
            max_rtn=max_rtn,
            row_id=row_id,
            customer_name=customer_name,
            cmc_filter_i=cmc_filter_i,
            py_filter_i=py_filter_i,
        )

    async def filter_array(self):
        mapper: AmherstMap = await mapper_from_query_csrname(self.csrname)
        fil_array = mapper.cmc_filters[self.cmc_filter_i].model_copy() if self.cmc_filter_i else FilterArray()

        if self.pk_value:
            fil_array.add_filter(FieldFilter(column=mapper.aliases.NAME, condition=self.condition, value=self.pk_value))

        if self.customer_name:
            if cust_con := mapper.connections.customer:
                customer_filter = FieldFilter(
                    column=CustomerAliases.CUSTOMER_NAME,
                    condition=self.condition,
                    value=self.customer_name,
                )
                fil_array.add_filter(ConnectedFieldFilter.from_fil(field_fil=customer_filter, connection=cust_con))
        return fil_array


class SearchResponse[T: AmherstShipableBase](BaseModel):
    records: list[T]
    length: int = 0
    search_request: SearchRequest
    more: MoreAvailable | None = None

    # @model_validator(mode='after')
    # def set_more(self):
    #     if self.more:
    #         self.more.html_link = ''
    #     return self

    def __str__(self):
        return (
            f'Search Response: {self.length}x {self.search_request.csrname if self.search_request.csrname else ', '.join(self.search_request.csrnames)} records'
            f'{' (' + str(self.more.n_more) + ' more available),' if self.more else '. '} '
            f'SearchRequest[{str(self.search_request)}]'
        )

    @model_validator(mode='after')
    def set_length(self):
        self.length = len(self.records)
        return self


class SearchResponseMulti(SearchResponse):
    search_request: Sequence[SearchRequest]

    def __str__(self):
        rtypes = '/'.join([req.csrname for req in self.search_request])
        return (
            f'Search Response with {self.length}x {rtypes} records. '
            f'SearchRequests[{'; '.join(str(_) for _ in self.search_request)}]'
            f'{', ' + str(self.more.n_more) + ' more available' if self.more else ''} '
        )


@dataclasses.dataclass
class MoreAvailableFront(MoreAvailable):
    json_link: str = None
    html_link: str = None
